/**
 * *******************************************************************************
 * Copyright (c) 2018-2022 Robert Bosch GmbH and others.
 * 
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Robert Bosch GmbH - initial API and implementation
 * *******************************************************************************
 */

package org.eclipse.app4mc.amalthea.visualization.hw.templates

import java.util.HashSet
import java.util.List
import org.eclipse.app4mc.amalthea.model.HwModule
import org.eclipse.app4mc.amalthea.model.HwStructure
import org.eclipse.app4mc.amalthea.model.INamed
import org.eclipse.app4mc.amalthea.model.ProcessingUnit
import org.eclipse.app4mc.visualization.util.svg.PlantUmlDiagram

class HWBlockGenerator {

	static def void updateDiagram(PlantUmlDiagram diagram, List<HwStructure> structures) {
		// reset old diagram data
		diagram.resetDiagramData();

		// generate new diagram
		val state = new State();
		val commands = new HashSet<String>();
		val result = generateStructures(structures, commands, state)

		if (state.success) {
			diagram.append(result);
		} else {
			// create error diagram 
			diagram.append(error(state.errorMessage))
		}
	}

	private static def error(String message) '''
		@startuml
		left to right direction
		actor User
		rectangle Visualization {
			usecase "Invalid Input" as ii
		}
		User ---> ii
		note top of ii
			
			«message»
			
		end note
		@enduml
	'''

	private static def CharSequence generateStructures(List<HwStructure> structures, HashSet<String> commands, State state) '''
			@startuml
			«IF !structures.empty»
				
				' ===== FRAMES =====
				«FOR s : structures»
					«generateLevels(s, commands, state)»
				«ENDFOR»
				
				' ===== ROUTES =====
				«FOR s : structures»
					«generateRoutes(s, commands, state)»
				«ENDFOR»
			«ELSE»
				«state.errorFlag = true»
				«state.errorMessage = " No Structure in the model. "»
			«ENDIF»
			
			skinparam component {
				backgroundColor<<ProcessingUnit>> #8CACFF
				borderColor<<ProcessingUnit>> #000000
				backgroundColor<<ConnectionHandler>> #FFFFA0
				borderColor<<ConnectionHandler>> #000000
				backgroundColor<<Memory>> #60FF82
				borderColor<<Memory>> #000000
				backgroundColor<<Cache>> #C8FFA6
				borderColor<<Cache>> #000000
				ArrowFontColor #C0C0C0
			}
			
			@enduml
		'''

	private static def CharSequence generateLevels(HwStructure s, HashSet<String> commands, State state) '''
		«IF !(s === null || s.name.nullOrEmpty)»
		
			frame "«s.name»" as «HWBlockGenerator.convertString(s.name)» {
			«IF !s.modules.empty»
			
				«FOR m: s.modules»
					«IF!(m.name.nullOrEmpty)»
						«addLevel(m, commands)»
					«ELSE»
						«state.errorFlag = true»
						«state.errorMessage = " Missing HW Module name. "»
					«ENDIF»
				«ENDFOR»
			«ENDIF»
			«FOR su : s.structures»
				«generateLevels(su, commands, state)»
			«ENDFOR»
		«ELSE»
			«state.errorFlag = true»
			«state.errorMessage = " Name of HwStructure is missing. "»
		«ENDIF»
			}
	'''

	private static def CharSequence addLevel(HwModule m, HashSet<String> commands) {
		val name = HWBlockGenerator.convertString(m.name)
		val command = "[" + name + "] as " + name + " <<" + m.eClass.name + ">>"

		if (commands.contains(command)) {
			// previously added -> do nothing
			""
		} else {
			commands.add(command)
			command
		}
	}

	private static def CharSequence generateRoutes(HwStructure s, HashSet<String> commands, State state) '''
		«IF !(s === null || s.connections.empty)»
			
			' add physical connections of structure «s.name»:
			«FOR c : s.connections»
				«»
					«addRoute(c.port1?.eContainer as INamed, c.port2?.eContainer as INamed, c, " -- ", commands)»
			«ENDFOR»
		«ENDIF»	
		«IF !(s === null || s.modules.empty)»
			«FOR m: s.modules»
				«IF m instanceof ProcessingUnit»
					«val pUnit = m as ProcessingUnit»
					«IF !pUnit.accessElements.empty»
						
						' add logical connections of processing unit «pUnit.name»:
						«FOR ae :pUnit.accessElements»
							«IF ae.destination !== null»
								«»
									«addRoute(pUnit, ae.destination, ae, " ..>", commands)»
							«ELSE»
								«state.errorFlag = true»
								«state.errorMessage = " Missing Destination of AccessElement. "»
							«ENDIF»	
						«ENDFOR»
					«ENDIF»
				«ENDIF»
			«ENDFOR»
		«ENDIF»
		«IF !(s === null || s.structures.empty)»
			«FOR su : s.structures»
				«generateRoutes(su, commands, state)»
			«ENDFOR»
		«ENDIF»
	'''

	private static def CharSequence addRoute(INamed source, INamed destination, INamed con, String conType, HashSet<String> commands) {
		val name1 = HWBlockGenerator.convertString(name(source))
		val name2 = HWBlockGenerator.convertString(name(destination))
		val conLabel = if(con.name.isNullOrEmpty) con.toString else con.name;
		val command = name1 + conType + name2 + " : " + conLabel

		if (commands.contains(command)) {
			// previously added -> do nothing
			""
		} else {
			commands.add(command)
			command
		}
	}

	private static def name(INamed obj) {
		if (obj === null) return "<undefined>"
		if (obj.name.nullOrEmpty) return "?"
		return obj.name
	}

	private static def convertString(String str) {
		return if (str.nullOrEmpty) "<no name>" else convertGenericString(str, "_");
	}

	private static def convertGenericString(String instring, String replacestring) {
		val String Regex = '/';
		val tmpstr0 = instring.replaceAll(Regex, replacestring);

		val String Regex1 = ' ';
		val tmpstr1 = tmpstr0.replaceAll(Regex1, replacestring);

		val String Regex2 = '\\(';
		val tmpstr2 = tmpstr1.replaceAll(Regex2, replacestring);

		val String Regex3 = '\\)';
		val tmpstr3 = tmpstr2.replaceAll(Regex3, replacestring);

		val String Regex4 = '-';
		val tmpstr4 = tmpstr3.replaceAll(Regex4, replacestring);

		val String Regex5 = '\\+';
		val tmpstr5 = tmpstr4.replaceAll(Regex5, replacestring);

		val String Regex6 = '%2f';
		val tmpstr6 = tmpstr5.replaceAll(Regex6, replacestring);

		val String Regex7 = '\\(';
		val tmpstr7 = tmpstr6.replaceAll(Regex7, replacestring);

		val String Regex8 = '\\)';
		val tmpstr8 = tmpstr7.replaceAll(Regex8, replacestring);

		val String Regex9 = '%2b';
		val String Replace9 = 'plus';
		val tmpstr9 = tmpstr8.replaceAll(Regex9, Replace9);

		val String Regex10 = '%3a%3a';
		val tmpstr10 = tmpstr9.replaceAll(Regex10, replacestring);

		val String Regex11 = '\\.';
		val tmpstr11 = tmpstr10.replaceAll(Regex11, replacestring);

		val String Regex12 = '#';
		val tmpstr12 = tmpstr11.replaceAll(Regex12, replacestring);

		val String Regex13 = '"';
		val tmpstr13 = tmpstr12.replaceAll(Regex13, replacestring);

		val String Regex14 = '$';
		val tmpstr14 = tmpstr13.replaceAll(Regex14, replacestring);

		val String Regex15 = '~';
		val tmpstr15 = tmpstr14.replaceAll(Regex15, replacestring);

		val String Regex16 = '%';
		val tmpstr16 = tmpstr15.replaceAll(Regex16, replacestring);

		val String Regex17 = '&';
		val tmpstr17 = tmpstr16.replaceAll(Regex17, replacestring);

		return tmpstr17;
	}

	private static class State {
		boolean errorFlag;
		String errorMessage;

		def State() {
			errorFlag = false;
		}

		def getErrorFlag() {
			return errorFlag;
		}

		def setErrorFlag(boolean flag) {
			errorFlag = flag;
		}

		def error() {
			return errorFlag;
		}

		def success() {
			return !errorFlag;
		}

		def String getErrorMessage() {
			return errorMessage;
		}

		def void setErrorMessage(String message) {
			errorMessage = message;
		}
	}

}
